#include "Board.h"
#include "common.h"
#include "dbgu.h"
#include "spi.h"
#include "sd_spi.h"
#include "fat/fs_tools.h"
#include "lcd/lcd3310.h"



void init(void)
{
	// enable peripheral clock for PIO controllers
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA);
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
	
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	//open debug channel
	debug_init_default();
	
	//initialize LEDs
	LED_R_INIT;
	LED_Y_INIT;
	LED_G_INIT;
	LED_B_INIT;
	
	//initialize SPI interface
	spiInit();
	
	//initialize LCD
	lcdInit();
	
	//initialize SD Card
	if(sdCardInit() == OK)
	{
		debug_msg("SD card initialized OK");
	}
	else
	{
		debug_msg("SD card initialization FAILED!");
	}
	
	//mount filesystem
	fsInit();
	
}


typedef struct __attribute__((packed))
{
	uint16_t bfType;		//magic "BM" string
	uint32_t bfSize;		//size of file in bytes
	uint16_t bfReserved1;	//zeros
	uint16_t bfReserved2;	//zeros
	uint32_t bfOffBits;		//at which byte data starts
} TBitmapHeader;

typedef struct __attribute__((packed))
{
	uint32_t biSize;		//size of this structure, usually 40
	uint32_t biWidth;		//width of the image
	uint32_t biHeight;		//height of the image
	uint16_t biPlanes;		//number of planes
	uint16_t biBitCount;	//number of bits per pixel
	uint32_t biCompression;	//type of compression
	uint32_t biSizeImage;	//size of image for compressed images
	uint32_t biXPelsPerMeter; //pixels per meter (not used, zero)
	uint32_t biYPelsPerMeter; //pixels per meter (not used, zero)
	uint32_t biClrUsed;		//number of colors used inbitmap (if zero use biBitCount)
	uint32_t clrImportant;	//numer of colors important for a bitmap. Use zero - all are important ;)
} TBitmapInfoHeader;


uint8_t lcdBitmapFromFile(char *fileName)
{
	FRESULT res;
	FIL bitmapFile;
	uint16_t bytesActuallyRead=0;
	TBitmapHeader bitmapHeader;
	TBitmapInfoHeader bitmapInfoHeader;
	
	//open file
	res = f_open(&bitmapFile, (const char*) fileName, FA_READ);
	if(res!=FR_OK) return(FAIL);
	
	//read bmp header
	res = f_read(&bitmapFile, &bitmapHeader, sizeof(TBitmapHeader), &bytesActuallyRead);
	if(res!=FR_OK) return(FAIL);
	if(bytesActuallyRead != sizeof(TBitmapHeader)) return(FAIL);
	
	//read bmp info header
	res = f_read(&bitmapFile, &bitmapInfoHeader, sizeof(TBitmapInfoHeader), &bytesActuallyRead);
	if(res!=FR_OK) return(FAIL);
	if(bytesActuallyRead != sizeof(TBitmapInfoHeader)) return(FAIL);
	
	//check bitmap dimensions
	if(bitmapInfoHeader.biWidth != LCD_X_SIZE) {debug_value("wrong X size =",bitmapInfoHeader.biWidth); return(FAIL);}
	if(bitmapInfoHeader.biHeight != LCD_Y_SIZE) {debug_value("wrong Y size =",bitmapInfoHeader.biHeight); return(FAIL);}
	
	//check color depth
	if(bitmapInfoHeader.biBitCount != 1) {debug_value("wrong color depth, bits/pixel =",bitmapInfoHeader.biBitCount); return(FAIL);}
	
	//set cursor at the beginning of bmp data
	res = f_lseek(&bitmapFile, bitmapHeader.bfOffBits);
	if(res!=FR_OK) return(FAIL);
	
	uint16_t bitInLine;
	uint8_t line;
	
	for(line = LCD_Y_SIZE; line > 0; --line)
	{
		uint8_t byteFromBitmap;
		
		//for pixels x=0 to x=83
		for(bitInLine = 0; bitInLine < LCD_X_SIZE; bitInLine++)
		{
			if( (bitInLine % 8) == 0)	//on every 8th bit...
			{
				//read next byte from bitmap file
				res = f_read(&bitmapFile, (void*) &byteFromBitmap, sizeof(uint8_t), &bytesActuallyRead);
			}
			
			if( (byteFromBitmap&(1<<7))==0 )	//if MSB of byte is clear
			{
				lcdPixelSet(bitInLine, line);	//this is "black" pixel on LCD
			}
			byteFromBitmap = byteFromBitmap << 1;	//shift byte
		}
		
		//dummy loops
		//bitmap data is aligned to 32 bits
		//e.g. Nokia 3310 display is 84 pixels wide. Next 32 bit word starts at 96th pixel
		//so there are 12 bits (1 byte and 4 bits) in .bmp file that are not displayed on screen
		//in most cases these bits are zeros
		const uint8_t DUMMY_BITS = 32 - LCD_X_SIZE % 32;
		for(bitInLine = LCD_X_SIZE; bitInLine < (LCD_X_SIZE+DUMMY_BITS); bitInLine++)
		{
			//read data...
			if( (bitInLine % 8) == 0)	//on every 8th bit...
			{
				//read next byte from bitmap file
				res = f_read(&bitmapFile, (void*) &byteFromBitmap, sizeof(uint8_t), &bytesActuallyRead);
			}
			//...but don't use it
		}
	}
	
	//close file
	res = f_close(&bitmapFile);
	if(res!=FR_OK) return(FAIL); else return(OK);
}


#define FILE_NAME_LEN		(8+1+3+1) //8 chars name, 1 char dot, 3 chars extension, 1 char NULL

int main( void )
{
	init();
	
	lcdSetFont(FONT_NORMAL);
	lcd("3310 Display",0,0);
	lcdSetFont(FONT_BOLD);
	lcd("SD+LCD Test",10,20);
	lcd("Press joystick",5,LCD_Y_SIZE-10);
	lcdRefresh();
	
	while( !( JOY_ENTER_PRESS || JOY_LEFT_PRESS || JOY_RIGHT_PRESS || JOY_UP_PRESS || JOY_DOWN_PRESS ) )
	{
		lcdInvertArea(0,LCD_X_SIZE,LCD_Y_SIZE-11,LCD_Y_SIZE-1);
		delay_ms(200);
		lcdRefresh();
	}
	
	lcdClear();
	lcdRefresh();
	
	//b/w bitmap movie
	uint8_t frame=0;
	const uint8_t NUMBER_OF_FRAMES = 6;
	char fileName[FILE_NAME_LEN] = "frame0.bmp";
	for(;;)
	{
		fileName[5] = '0' + frame;
		lcdClear();
		if(lcdBitmapFromFile(fileName)!=OK)
		{
			debug_msg("lcdBitmapFromFile error on "); debug_txt(fileName);
		}
		lcdRefresh();
		if(frame < NUMBER_OF_FRAMES) frame++; else frame = 0;
		delay_ms(100);
		
	}
	
	return(0);
}


