#include "common.h"
#include "lcd_2x16.h"
#include "adc.h"


//Clock selection for Timer
#define TC_CLKS_MCK2			0x0
#define TC_CLKS_MCK8			0x1
#define TC_CLKS_MCK32			0x2
#define TC_CLKS_MCK128			0x3
#define TC_CLKS_MCK1024			0x4
#define SELECTED_TC_CLKS		TC_CLKS_MCK128
#define TC_PRESCALER			128

#define TIMER					AT91C_BASE_TC0
#define TIMER_ID				AT91C_ID_TC0
#define ADC						AT91C_BASE_ADC


void adcIrqHandler(void);
void adcTimerInit(void);


void adcSamplingInit(uint8_t channel)
{
    #if ADC_RESOLUTION_10_BIT==1
	// set ADC mode
	ADC->ADC_MR =
		AT91C_ADC_TRGEN_EN |
		AT91C_ADC_TRGSEL_TIOA0 |
		(0x4<<8) |		// max prescale setting
		(0x0C<<16) |	// min startup setting
		(0x9<<24);		// min sample-and-hold time
								
    #else
	ADC->ADC_MR =
		AT91C_ADC_TRGEN_EN |
		AT91C_ADC_TRGSEL_TIOA0 |
		AT91C_ADC_LOWRES_8_BIT |
		(0x2<<8) |		// prescale setting, fACD=8MHz=fmax
		(0x14<<16) |	// startup setting
		(0xB<<24);		// sample-and-hold time (>1470ns for 10k)
    #endif
	
	//initialize Timer
    adcTimerInit();

	//mask channel (prevent forbidden values)
    channel = channel & 0x0F;
	
	//disable all registers - to be sure
    ADC->ADC_CHDR =	0xF;
	
	//enable selected channel
    ADC->ADC_CHER =	1<<channel;
    
	//define ADC's IRQ priority
	const uint8_t ADC_INTERRUPT_PRIORITY = 2;
	
	//prepare AIC for ADC interrupts
	AT91F_AIC_ConfigureIt ( AT91C_BASE_AIC, AT91C_ID_ADC, ADC_INTERRUPT_PRIORITY, AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL, adcIrqHandler);
	
	//enable interrupt from ADC in AIC
	AT91F_AIC_EnableIt (AT91C_BASE_AIC, AT91C_ID_ADC);
	
	//enable ADC interrupt on DRDY bit
	ADC->ADC_IER = AT91C_ADC_DRDY;
}




void adcTimerInit(void)
{
    unsigned int dummy;
    const uint32_t INITIAL_SAMPLING_RATE = 500;	//[samples per second]

    // First, enable the clock of the TIMER
    AT91C_BASE_PMC->PMC_PCER = 1<<TIMER_ID;
	
	//disable timer's clock
	TIMER->TC_CCR = AT91C_TC_CLKDIS;
	
	//disable timer's interrupts
	TIMER->TC_IDR = 0xFFFFFFFF ;
	
	//read status register to clear its flags
	dummy = TIMER->TC_SR;
	dummy = dummy;
	
	//set capture mode register of the timer
	TIMER->TC_CMR = AT91C_TC_WAVE  //bit WAVE=1 => Waveform Mode
		|SELECTED_TC_CLKS   //selected Timer clock freq.
		|AT91C_TC_ACPA_SET  //RA Compare Effect on TIOA: set (01)
		|AT91C_TC_ACPC_CLEAR  //RC Compare Effect on TIOA: clear (10)
		|AT91C_TC_WAVESEL_UP_AUTO;  //Waveform Selection: UP mode with automatic trigger on RC compare (10)
	
	adcSetSamplingRate(INITIAL_SAMPLING_RATE);
	
    //enable timer's clock
	TIMER->TC_CCR = AT91C_TC_CLKEN;
	
    //start timer
    TIMER->TC_CCR = AT91C_TC_SWTRG ;
}




uint8_t adcSetSamplingRate(uint32_t sps)
{
	if(sps<SAMPLING_RATE_MIN) return(0);
	if(sps>SAMPLING_RATE_MAX) return(0);
	
	//compute value for timer full-count
	uint16_t TC_RCValue = MCK/TC_PRESCALER/sps;
	
	//prevent zero divider when using TC_RCValue / 2
	if(sps < 2) return(0);
	
	//write full-count value into RC
	TIMER->TC_RC = TC_RCValue;
	
	//write 0.5 of full value to RA
	TIMER->TC_RA = TC_RCValue/2;
	
	return(1);
}


void adcStopSampling(void)
{
	//AT91F_PMC_DisablePeriphClock ( AT91C_BASE_PMC, 1<< AT91C_ID_TC0 ) ;
	AT91C_BASE_PMC->PMC_PCER = TIMER_ID;
}


void adcStartSampling(void)
{
//	AT91F_PMC_EnablePeriphClock ( AT91C_BASE_PMC, 1<< AT91C_ID_TC0 ) ;
	AT91C_BASE_PMC->PMC_PCDR = TIMER_ID;
}






uint16_t adcGet(uint8_t channel)
{
	//define configuration fields in ADC_MR register
	const uint32_t SHTIM_MASK = 0x00 << 24;   //sample and hold time
	const uint32_t STARTUP_MASK = 0x04 << 16; //startup time
	const uint32_t PRESCAL_MASK = 0x4 << 8;   //prescaler setting
	
	//configure ADC
	ADC->ADC_MR = SHTIM_MASK | STARTUP_MASK | PRESCAL_MASK | AT91C_ADC_TRGEN_EN;
	
	//activate requested channel
    channel = channel & 0x07;
    ADC->ADC_CHER =	1<<channel;
	
	//start conversion!
	ADC->ADC_CR = AT91C_ADC_START;
	
	//wait for adc DRDY bit set
	while( ! (ADC->ADC_SR & AT91C_ADC_DRDY) );
	
	//read result from LCDR register
	uint16_t retval = ADC->ADC_LCDR;
	
	//disable all channels
	ADC->ADC_CHDR =	0xFF;
	
	return(retval);
}



void adcIrqHandler(void)
{
	LED_R_TOGGLE;
	uint16_t adcResult;
	adcResult = ADC->ADC_LCDR;
	lcdLocate(1,6);
	lcdValue(adcResult);
	lcdText("  ");
}




