#include "common.h"
#include "lcd_2x16.h"
#include "adc.h"

#define ADC_BUFFER_SIZE		40	//in 16-bit words and doubled!

//buffers for samples
static volatile int16_t adcBuffer[2][ADC_BUFFER_SIZE];
static volatile uint8_t activeBuffer;

//prototype for ADC DMA ISR function
void adcDmaBufferSwap(void);


void init(void)
{
	// enable peripheral clock for PIO controllers
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA);
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	debug_init_default();
	debug_msg("debug channel initialized");
	
	//initialize LEDs
	LED_R_INIT;
	LED_Y_INIT;
	LED_G_INIT;
	LED_B_INIT;
	
	lcdInit();
	lcdLocate(0,0);
	lcdText("max:");
}




int main( void )
{
	init();
	
	//define channel to sample
	const uint8_t ADC_VR_CHANNEL = 7;
	const uint32_t ADC_DEFAULT_SAMPLING_RATE = 200;	//samples per second
	
	//start DMA transfers from ADC
	adcDmaStart(ADC_VR_CHANNEL, ADC_DEFAULT_SAMPLING_RATE, adcDmaBufferSwap);
	
	for(;;)
	{
		LED_B_TOGGLE;
		delay_ms(50);
	}
	return(0);
	
}



//this function is called from adc.c module on ADC's RXBUFF and ENDRX flags
void adcDmaBufferSwap(void)
{

	//debug_msg("ISR");
	
	//copy adc status to a variable
	static uint32_t status;
	status = ADC->ADC_SR;
	
	//both buffers are not initialized or full
	if(status & AT91C_ADC_RXBUFF)
	{
		debug_msg("DMA Init");
		activeBuffer = 0;
		ADC->ADC_RPR = (uint32_t) &adcBuffer[0][0];
		ADC->ADC_RNPR = (uint32_t) &adcBuffer[1][0];
		ADC->ADC_RCR = ADC_BUFFER_SIZE;
		ADC->ADC_RNCR = ADC_BUFFER_SIZE;
		ADC->ADC_PTCR = AT91C_PDC_RXTEN;
		return;
	}
	
	//buffer filling finished
	if(status & AT91C_ADC_ENDRX)
	{
		LED_G_OFF;
		
		static uint8_t bufferToProcess;
		if(activeBuffer)
		{
			//debug_msg("RXBUFF 0");
			activeBuffer = 0;
			bufferToProcess = 1;
		}
		else
		{
			//debug_msg("RXBUFF 1");
			activeBuffer = 1;
			bufferToProcess = 0;
		}
		//prepare "next" buffer
		ADC->ADC_RNPR = (uint32_t) &adcBuffer[activeBuffer][0];
		ADC->ADC_RNCR = ADC_BUFFER_SIZE;
		
		//this is a good place to copy/process/display data in buffer that has just been rxed
		static uint16_t i;
		static int16_t max = 0;
		static int32_t mean = 0;
		
		//shift ADC data to zero mean value
		for(i=0;i<ADC_BUFFER_SIZE;i++)
		{
			//shift samples to zero DC offset
			adcBuffer[bufferToProcess][i] -= 512; //0.5 of full scale, 1024
			
			//compute absolute value
			if(adcBuffer[bufferToProcess][i] < 0) adcBuffer[bufferToProcess][i] = -adcBuffer[bufferToProcess][i];
			
			//compute max sample value
			if(adcBuffer[bufferToProcess][i] > max) max = adcBuffer[bufferToProcess][i];
			
			mean += adcBuffer[bufferToProcess][i];
		}
		
		max = max / 32 + 1;
		
		lcdLocate(1,0);
		for(i=0;i<max;i++) lcdChr('#');
		for(i=max;i<16;i++) lcdChr('-');
		
		LED_G_ON;
		
	}
	
}


