#include "common.h"
#include "msc_usb.h"
#include "msc_scsi.h"
#include "msc_if.h"



#define SCSI_INQUIRY                        0x12    ///< Inquiry Command
#define SCSI_READ_CAPACITY                  0x25    ///< Read Capacity Command
#define SCSI_READ                           0x28    ///< Read Command
#define SCSI_WRITE                          0x2A    ///< Write Command
#define SCSI_MODE_SENSE                     0x1A    ///< Mode Sense Command
#define SCSI_REQUEST_SENSE                  0x03    ///< Request Sense Command
#define SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL   0x1E    ///< Prevent/Allow Medium Removal Command
#define SCSI_TEST_UNIT_READY                0x00    ///< Test Unit Ready Command
#define SCSI_VERIFY                         0x2F    ///< Verify Command - not supported here
#define SCSI_START_STOP                     0x1B    ///< Start/Stop Medium Command
#define SCSI_READ_FORMAT_CAPACITIES			0x23	//LOOK FOR THIS IN Multimedia Commands spec!


#define S_NO_SENSE 0x0
#define S_RECOVERED_ERROR 0x1
#define S_NOT_READY 0x2
#define S_MEDIUM_ERROR 0x3
#define S_HARDWARE_ERROR 0X4
#define S_ILLEGAL_REQUEST 0x5
#define S_UNIT_ATTENTION 0x6
#define S_DATA_PROTECT 0x7
#define S_BLANK_CHECK 0x8
#define S_VENDOR_SPECIFIC 0x9
#define S_COPY_ABORTED 0xa
#define S_ABORTED_COMMAND 0xb
#define S_OBSOLETE 0xc
#define S_VOLUME_OVERFLOW 0xd
#define S_MISCOMPARE 0xe

#define S_CURRENT 0x70
#define S_DEFERRED 0x71

// ASC ASCQ Codes for Sense Data (only those that we plan to use)
// with sense key Illegal request for a command not supported
#define ASC_INVALID_COMMAND_OPCODE 0x20
#define ASCQ_INVALID_COMMAND_OPCODE 0x00

// from SPC-3 Table 185
// with sense key Illegal Request for test unit ready
#define ASC_LOGICAL_UNIT_NOT_SUPPORTED 0x25
#define ASCQ_LOGICAL_UNIT_NOT_SUPPORTED 0x00

// with sense key Not ready
#define ASC_LOGICAL_UNIT_DOES_NOT_RESPOND 0x05
#define ASCQ_LOGICAL_UNIT_DOES_NOT_RESPOND 0x00

#define ASC_MEDIUM_NOT_PRESENT 0x3a
#define ASCQ_MEDIUM_NOT_PRESENT 0x00

#define ASC_LOGICAL_UNIT_NOT_READY_CAUSE_NOT_REPORTABLE 0x04
#define ASCQ_LOGICAL_UNIT_NOT_READY_CAUSE_NOT_REPORTABLE 0x00

#define ASC_LOGICAL_UNIT_IN_PROCESS 0x04
#define ASCQ_LOGICAL_UNIT_IN_PROCESS 0x01

#define ASC_LOGICAL_UNIT_NOT_READY_INIT_REQD 0x04
#define ASCQ_LOGICAL_UNIT_NOT_READY_INIT_REQD 0x02

#define ASC_LOGICAL_UNIT_NOT_READY_INTERVENTION_REQD 0x04
#define ASCQ_LOGICAL_UNIT_NOT_READY_INTERVENTION_REQD 0x03

#define ASC_LOGICAL_UNIT_NOT_READY_FORMATTING 0x04
#define ASCQ_LOGICAL_UNIT_NOT_READY_FORMATTING 0x04

#define ASC_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE 0x21
#define ASCQ_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE 0x00

#define ASC_WRITE_PROTECTED 0x27
#define ASCQ_WRITE_PROTECTED 0x00


typedef struct
{
    uint8_t		ResponseCode;					//byte0
    uint8_t		Obsolete;						//byte1
    uint8_t		SenseKey;						//byte2
    uint32_t 	Information;					//byte3..6
	uint8_t		AdditionalSenseLength;			//byte7
	uint32_t	CommandSpecificInformation;		//byte8..11
	uint8_t		ASC; 							//byte12, Additional Sense Code
	uint8_t		ASCQ;							//byte13, Additional Sense Code Qualifier
	uint8_t		FieldReplaceableUnitCode;		//byte14
	uint8_t		SenseKeySpecific[3];			//byte15..17
}T_scsiSenseData;



#define MSC_BUF_SIZE						0x080
#define MSC_DATA_BUF_SIZE				(1024)
#define DEFAULT_SECTOR_SIZE			512			//bytes
#define BUF_OVERHEAD						0

static volatile uint16_t mscSectorSize=DEFAULT_SECTOR_SIZE;
static volatile uint16_t mscDataBufSizeSectors=MSC_DATA_BUF_SIZE/DEFAULT_SECTOR_SIZE;

extern const uint8_t scsiInquiryResponse[];

uint8_t scsiProcessInquiry(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessRequestSense(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessModeSense(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessReadCapacity(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessReadFormatCapacities(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessRead(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessTestUnitReady(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessVerify(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessWrite(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiProcessPreventAllowMediumRemoval(T_Cbw* pCBW, uint32_t *dataProcessed);
uint8_t scsiInitSenseData(T_Cbw* pCBW);


volatile uint8_t mscBuf[MSC_BUF_SIZE];
volatile uint32_t mscActualDataBuf[MSC_DATA_BUF_SIZE/4+BUF_OVERHEAD];
volatile uint8_t *mscDataBuf = (uint8_t*)&mscActualDataBuf;
volatile T_scsiSenseData scsiSense;
volatile T_scsiSenseData *pscsiSense=&scsiSense;



//processes SCSI Command Block
//alters processedDataLength -- number of bytes processed
//returns processing status
uint8_t scsiProcess(T_Cbw *pCBW, uint32_t *processedDataLength)
{
	*processedDataLength=0;
	switch(pCBW->bCBWCB[0])
	{
		case SCSI_INQUIRY:
		{
			return(scsiProcessInquiry(pCBW,processedDataLength));
		}
		case SCSI_REQUEST_SENSE:
		{
			return(scsiProcessRequestSense(pCBW,processedDataLength));
		}
		case SCSI_MODE_SENSE:
		{
			return(scsiProcessModeSense(pCBW,processedDataLength));
		}
		case SCSI_READ_CAPACITY:
		{
			return(scsiProcessReadCapacity(pCBW,processedDataLength));
		}
		case SCSI_READ_FORMAT_CAPACITIES:
		{
			return(scsiProcessReadFormatCapacities(pCBW,processedDataLength));
		}
		case SCSI_READ:
		{
			return(scsiProcessRead(pCBW,processedDataLength));
		}
		case SCSI_TEST_UNIT_READY:
		{
			return(scsiProcessTestUnitReady(pCBW,processedDataLength));
		}
		case SCSI_VERIFY:
		{
			return(scsiProcessVerify(pCBW,processedDataLength));
		}
		case SCSI_WRITE:
		{
			return(scsiProcessWrite(pCBW,processedDataLength));
		}
		case SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL:
		{
			return(scsiProcessPreventAllowMediumRemoval(pCBW,processedDataLength));
		}
		default:
		{
			return(1);
		}
	}
	return(0);
}



uint8_t scsiProcessPreventAllowMediumRemoval(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	*dataProcessed=0;
	return(0);
}



uint8_t scsiProcessVerify(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	*dataProcessed=0;
	return(0);
}



uint8_t scsiProcessWrite(T_Cbw* pCBW, uint32_t *dataProcessed)
{

    uint32_t writeLBA=0;
    writeLBA=( pCBW->bCBWCB[2]<<(3*8) )|
            ( pCBW->bCBWCB[3]<<(2*8) )|
            ( pCBW->bCBWCB[4]<<(1*8) )|
            ( pCBW->bCBWCB[5]<<(0*8) );
    uint16_t TransferLength_blocks=( pCBW->bCBWCB[7]<<8 )|( pCBW->bCBWCB[8] );
    uint32_t TransferLength_bytes=TransferLength_blocks<<9; //blocks*512B
    uint16_t BlockByteCount=0;
    uint32_t TotalByteCount=0;
    uint16_t LBACount=0;
    uint16_t temp;
	uint16_t bufCounter=0;
	uint8_t retval=0;
	
    for(;;)
    {
	    if (AT91C_BASE_UDP->UDP_ISR & AT91C_UDP_EPINT2)
        {
		    AT91C_BASE_UDP->UDP_ICR = AT91C_UDP_EPINT2;
            
            if (AT91C_BASE_UDP->UDP_CSR[2] & USB_EP2_CurrentBank)
            {
                temp=USB_DataOut((uint8_t *)(&mscDataBuf[bufCounter]));
                BlockByteCount+=temp;
				bufCounter+=temp;
                TotalByteCount+=temp;
                
                if(BlockByteCount>=mscSectorSize)
                {
                    BlockByteCount=0;
					LBACount++;
					if(bufCounter>=MSC_DATA_BUF_SIZE)
					{
						if(mscWriteSectors(writeLBA,LBACount,(void*)(mscDataBuf),pCBW->bCBWLUN)==0) retval=1;
						bufCounter=0;
						writeLBA+=LBACount;
						LBACount=0;
					}
                }
                
                if(TotalByteCount>=TransferLength_bytes)
                {
                    *dataProcessed=TransferLength_bytes;
					if(LBACount==0) return(retval);
					if(mscWriteSectors(writeLBA,LBACount,(void*)(mscDataBuf),pCBW->bCBWLUN)==0) retval=1;
					if(retval)
					{
						pscsiSense->SenseKey = S_MEDIUM_ERROR;
					}
					return(retval);
                }
                
            }//if (AT91C_BASE_UDP->UDP_CSR[2] & USB_EP2_CurrentBank)
        } // if (Status & AT91C_UDP_EPINT2)
    }//for(;;)
	
	if(retval)
	{
		pscsiSense->SenseKey = S_MEDIUM_ERROR;
	}
	return(retval);
}





uint8_t scsiProcessTestUnitReady(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	*dataProcessed=0;
	//set the flags
	return(0);
}







uint8_t scsiProcessRead(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	*dataProcessed=0;

    uint32_t ReadLBA=0;
    ReadLBA=( pCBW->bCBWCB[2]<<(3*8) )|
            ( pCBW->bCBWCB[3]<<(2*8) )|
            ( pCBW->bCBWCB[4]<<(1*8) )|
            ( pCBW->bCBWCB[5]<<(0*8) );
    uint16_t dataRemaining=( pCBW->bCBWCB[7]<<8 )|( pCBW->bCBWCB[8] );
	*dataProcessed=mscSectorSize*dataRemaining;
	uint8_t retval=0;

	while(dataRemaining)
	{
		if(dataRemaining>mscDataBufSizeSectors)
		{
			dataRemaining-=mscDataBufSizeSectors;
			if(mscReadSectors(ReadLBA,mscDataBufSizeSectors,(void*)mscDataBuf,pCBW->bCBWLUN)==0) retval=1;
			USB_DataIn((uint8_t *)mscDataBuf,MSC_DATA_BUF_SIZE);
			ReadLBA+=mscDataBufSizeSectors;
		}
		else
		{
			if(mscReadSectors(ReadLBA,dataRemaining,(void*)mscDataBuf,pCBW->bCBWLUN)==0) retval=1;
			USB_DataIn((uint8_t *)mscDataBuf,dataRemaining*mscSectorSize);
			if(retval)
			{
				pscsiSense->SenseKey = S_MEDIUM_ERROR;
			}
			return(retval);
		}
	}
	
	if(retval)
	{
		pscsiSense->SenseKey = S_MEDIUM_ERROR;
	}
	return(retval);
	
}



uint8_t scsiProcessReadCapacity(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	uint32_t diskSize;
	uint16_t sectorSize;
	
	if(mscGetCapacity(&diskSize,&sectorSize,pCBW->bCBWLUN))
	{
		mscBuf[0x00]=(diskSize>>(3*8));
		mscBuf[0x01]=(diskSize>>(2*8));
		mscBuf[0x02]=(diskSize>>(1*8));
		mscBuf[0x03]=(diskSize>>(0*8));
		mscBuf[0x04]=(sectorSize>>(3*8));
		mscBuf[0x05]=(sectorSize>>(2*8));
		mscBuf[0x06]=(sectorSize>>(1*8));
		mscBuf[0x07]=(sectorSize>>(0*8));
		*dataProcessed = 8;
		USB_DataIn((uint8_t *)mscBuf,(*dataProcessed));
		mscDataBufSizeSectors=MSC_DATA_BUF_SIZE/sectorSize; //update global variable
		mscSectorSize=sectorSize; //update global variable
		return(0);
	}
	else
	{
		return(1);
	}
}



uint8_t scsiProcessReadFormatCapacities(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	
	uint16_t allocationLength = (pCBW->bCBWCB[7]<<8) | (pCBW->bCBWCB[8]);

    uint32_t NumberOfBlocks=0;
    uint16_t BlockLength=0;
    mscGetCapacity(&NumberOfBlocks,&BlockLength,pCBW->bCBWLUN);
	mscSectorSize=BlockLength; //update global variable
	mscDataBufSizeSectors=MSC_DATA_BUF_SIZE/BlockLength;	//update global variable
	
	mscBuf[0]=0x00;
	mscBuf[1]=0x00;
	mscBuf[2]=0x00;
	mscBuf[3]=0x08;
    mscBuf[4]=(uint8_t)(NumberOfBlocks>>(3*8));
    mscBuf[5]=(uint8_t)(NumberOfBlocks>>(2*8));
    mscBuf[6]=(uint8_t)(NumberOfBlocks>>(1*8));
    mscBuf[7]=(uint8_t)(NumberOfBlocks>>(0*8));
	mscBuf[8]=0x02<<2;
    mscBuf[9]=(uint8_t)(BlockLength>>(2*8));
    mscBuf[10]=(uint8_t)(BlockLength>>(1*8));
    mscBuf[11]=(uint8_t)(BlockLength>>(0*8));
    mscBuf[12]=(uint8_t)(NumberOfBlocks>>(3*8));
    mscBuf[13]=(uint8_t)(NumberOfBlocks>>(2*8));
    mscBuf[14]=(uint8_t)(NumberOfBlocks>>(1*8));
    mscBuf[15]=(uint8_t)(NumberOfBlocks>>(0*8));
	mscBuf[16]=0x00;
    mscBuf[17]=(uint8_t)(BlockLength>>(2*8));
    mscBuf[18]=(uint8_t)(BlockLength>>(1*8));
    mscBuf[19]=(uint8_t)(BlockLength>>(0*8));
	
	*dataProcessed = 20;
	USB_DataIn((uint8_t *)mscBuf,allocationLength); //host requests more bytes, but will get only 20
	return(0);
}



uint8_t scsiInitSenseData(T_Cbw* pCBW)
{
	pscsiSense->ResponseCode = 0x70;
	pscsiSense->Obsolete = 0x00;
	pscsiSense->SenseKey = S_NO_SENSE;
	pscsiSense->Information = 0x00;
	pscsiSense->AdditionalSenseLength = 0x0A;
	pscsiSense->CommandSpecificInformation = 0x00;
	pscsiSense->ASC = 0x00;
	pscsiSense->ASCQ = 0x00;
	pscsiSense->FieldReplaceableUnitCode = 0x00;
	pscsiSense->SenseKeySpecific[0] = 0x00;
	pscsiSense->SenseKeySpecific[1] = 0x00;
	pscsiSense->SenseKeySpecific[2] = 0x00;
	return(1);
}



uint8_t scsiProcessRequestSense(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	*dataProcessed = pCBW->bCBWCB[4];
	USB_DataIn((void*)pscsiSense,(*dataProcessed));
	scsiInitSenseData(pCBW);
	return(0);
}






uint8_t scsiProcessModeSense(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	uint8_t allocationLength = pCBW->bCBWCB[4];
	
	mscBuf[0x00]=0x03;
	mscBuf[0x01]=0x00;
	mscBuf[0x02]=0x00;
	mscBuf[0x03]=0x00;
	*dataProcessed=4;
	USB_DataIn((uint8_t *)mscBuf,allocationLength);
	return(0);
}



uint8_t scsiProcessInquiry(T_Cbw* pCBW, uint32_t *dataProcessed)
{
	uint16_t length = pCBW->bCBWCB[3]<<8 | pCBW->bCBWCB[4];
	USB_DataIn((uint8_t *)scsiInquiryResponse,length);
	scsiInitSenseData(pCBW);
	*dataProcessed = length;
	return(0);
}




const uint8_t scsiInquiryResponse[]=
{
    0x00,           //00, PERIPHERAL QUALIFIER, PERIPHERAL DEVICE TYPE: device is connected to logical unit, direct access device type
    0x80,           //01, RMB (bit7 set to 1) => removable medium
    0x04,           //02, The device complies to ANSI INCITS 351-2001 (SPC-2)
    0x02,           //03, According to SPC-3: >2 reserved, <2 obsolete => Response data format=2
    0x20,           //04, Additional Length
    0x00,           
    0x00,
    0x00,
    //T10 Vendor ID
    'R',
    'A',
    'B',
    'W',
    ' ',
    ' ',
    ' ',
    ' ',
    //Product ID
    'S',
    'A',
    'M',
    '7',
    'X',
    'C',
    ' ',
    'E',
    'v',
    'a',
    'l',
    ' ',
    ' ',
    ' ',
    ' ',
    ' ',
    //Product Revision Level
    '0',
    '0',
    '0',
    '0'
};





