//*----------------------------------------------------------------------------
//*      ATMEL Microcontroller Software Support  -  ROUSSET  -
//*----------------------------------------------------------------------------
//* The software is delivered "AS IS" without warranty or condition of any
//* kind, either express, implied or statutory. This includes without
//* limitation any warranty or condition with respect to merchantability or
//* fitness for any particular purpose, or against the infringements of
//* intellectual property rights of others.
//*----------------------------------------------------------------------------
//* File Name           : cdc_enumerate.c
//* Object              : Handle CDC enumeration
//*
//* 1.0 Apr 20 200 	: ODi Creation
//* 1.1 14/Sep/2004 JPP : Minor change
//* 1.1 15/12/2004  JPP : suppress warning
//* 1.2 30-Jun-2006 JPP : Set in AT91C_UDP_EPTYPE_INT_IN
//* 3-July-2008 RABW    : Simplified and modified to fit AT91SAM7XC128/256 eval kit
//*                       Original source downloaded from:
//*                       http://www.siwawi.arubi.uni-kl.de/avr_projects/arm_projects/at91sam7s_usart_usb_20061027.zip
//*----------------------------------------------------------------------------
#include "common.h"
#include "cdc_enumerate.h"

//configure USB DP pullup control pin location
#define CDC_USB_PULLUP_PIO		AT91C_BASE_PIOA
#define CDC_USB_PULLUP_MASK		AT91C_PIO_PA9

//configure VDETECT pin location
#define CDC_USB_VDET_PIO		AT91C_BASE_PIOA
#define CDC_USB_VDET_MASK		AT91C_PIO_PA8

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

#define AT91C_EP_IN_SIZE 0x40

const uint8_t devDescriptor[] = {
	/* Device descriptor */
	0x12,   // bLength
	0x01,   // bDescriptorType
	0x10,   // bcdUSBL
	0x01,   //
	0x02,   // bDeviceClass:    CDC class code
	0x00,   // bDeviceSubclass: CDC class sub code
	0x00,   // bDeviceProtocol: CDC Device protocol
	0x08,   // bMaxPacketSize0
	0xEB,   // idVendorL
	0x03,   //
	// mtmt 0x24,   // idProductL
	0x25,   // idProductL
	0x61,   //
	0x10,   // bcdDeviceL
	0x01,   //
	0x00,   // iManufacturer    // 0x01
	0x00,   // iProduct
	0x00,   // SerialNumber
	0x01    // bNumConfigs
};

const uint8_t cfgDescriptor[] = {
	/* ============== CONFIGURATION 1 =========== */
	/* Configuration 1 descriptor */
	0x09,   // CbLength
	0x02,   // CbDescriptorType
	0x43,   // CwTotalLength 2 EP + Control
	0x00,
	0x02,   // CbNumInterfaces
	0x01,   // CbConfigurationValue
	0x00,   // CiConfiguration
	0xC0,   // CbmAttributes 0xA0
	0x00,   // CMaxPower

	/* Communication Class Interface Descriptor Requirement */
	0x09, // bLength
	0x04, // bDescriptorType
	0x00, // bInterfaceNumber
	0x00, // bAlternateSetting
	0x01, // bNumEndpoints
	0x02, // bInterfaceClass
	0x02, // bInterfaceSubclass
	0x00, // bInterfaceProtocol
	0x00, // iInterface

	/* Header Functional Descriptor */
	0x05, // bFunction Length
	0x24, // bDescriptor type: CS_INTERFACE
	0x00, // bDescriptor subtype: Header Func Desc
	0x10, // bcdCDC:1.1
	0x01,

	/* ACM Functional Descriptor */
	0x04, // bFunctionLength
	0x24, // bDescriptor Type: CS_INTERFACE
	0x02, // bDescriptor Subtype: ACM Func Desc
	0x00, // bmCapabilities

	/* Union Functional Descriptor */
	0x05, // bFunctionLength
	0x24, // bDescriptorType: CS_INTERFACE
	0x06, // bDescriptor Subtype: Union Func Desc
	0x00, // bMasterInterface: Communication Class Interface
	0x01, // bSlaveInterface0: Data Class Interface

	/* Call Management Functional Descriptor */
	0x05, // bFunctionLength
	0x24, // bDescriptor Type: CS_INTERFACE
	0x01, // bDescriptor Subtype: Call Management Func Desc
	0x00, // bmCapabilities: D1 + D0
	0x01, // bDataInterface: Data Class Interface 1

	/* Endpoint 1 descriptor */
	0x07,   // bLength
	0x05,   // bDescriptorType
	0x83,   // bEndpointAddress, Endpoint 03 - IN
	0x03,   // bmAttributes      INT
	0x08,   // wMaxPacketSize
	0x00,
	0xFF,   // bInterval

	/* Data Class Interface Descriptor Requirement */
	0x09, // bLength
	0x04, // bDescriptorType
	0x01, // bInterfaceNumber
	0x00, // bAlternateSetting
	0x02, // bNumEndpoints
	0x0A, // bInterfaceClass
	0x00, // bInterfaceSubclass
	0x00, // bInterfaceProtocol
	0x00, // iInterface

	/* First alternate setting */
	/* Endpoint 1 descriptor */
	0x07,   // bLength
	0x05,   // bDescriptorType
	0x01,   // bEndpointAddress, Endpoint 01 - OUT
	0x02,   // bmAttributes      BULK
	AT91C_EP_OUT_SIZE,   // wMaxPacketSize
	0x00,
	0x00,   // bInterval

	/* Endpoint 2 descriptor */
	0x07,   // bLength
	0x05,   // bDescriptorType
	0x82,   // bEndpointAddress, Endpoint 02 - IN
	0x02,   // bmAttributes      BULK
	AT91C_EP_IN_SIZE,   // wMaxPacketSize
	0x00,
	0x00    // bInterval
};

/* USB standard request code */
#define STD_GET_STATUS_ZERO           0x0080
#define STD_GET_STATUS_INTERFACE      0x0081
#define STD_GET_STATUS_ENDPOINT       0x0082

#define STD_CLEAR_FEATURE_ZERO        0x0100
#define STD_CLEAR_FEATURE_INTERFACE   0x0101
#define STD_CLEAR_FEATURE_ENDPOINT    0x0102

#define STD_SET_FEATURE_ZERO          0x0300
#define STD_SET_FEATURE_INTERFACE     0x0301
#define STD_SET_FEATURE_ENDPOINT      0x0302

#define STD_SET_ADDRESS               0x0500
#define STD_GET_DESCRIPTOR            0x0680
#define STD_SET_DESCRIPTOR            0x0700
#define STD_GET_CONFIGURATION         0x0880
#define STD_SET_CONFIGURATION         0x0900
#define STD_GET_INTERFACE             0x0A81
#define STD_SET_INTERFACE             0x0B01
#define STD_SYNCH_FRAME               0x0C82

/* CDC Class Specific Request Code */
#define GET_LINE_CODING               0x21A1
#define SET_LINE_CODING               0x2021
#define SET_CONTROL_LINE_STATE        0x2221


AT91S_CDC_LINE_CODING line = {
	115200, // baudrate
	0,      // 1 Stop Bit
	0,      // None Parity
	8
};     // 8 Data bits

// mtmt uint32_t currentReceiveBank = AT91C_UDP_RX_DATA_BK0;

uint8_t currentConfiguration;
uint8_t currentConnection;
unsigned int  currentRcvBank;
uint8_t lineCodingChanged=0;



static AT91PS_UDP pUdp;
static uint8_t AT91F_UDP_IsConfigured();
static void AT91F_CDC_Enumerate();

//returns 1 if line coding has changed since last call
uint8_t cdcGetLineCoding(AT91PS_CDC_LINE_CODING lineCoding)
{
	if(lineCodingChanged)
	{
		lineCoding->dwDTERRate = line.dwDTERRate;
		lineCoding->bCharFormat = line.bCharFormat;
		lineCoding->bParityType = line.bParityType;
		lineCoding->bDataBits = line.bDataBits;
		lineCodingChanged = 0;
		return(1);
	}
	return(0);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_CDC_Open
//* \brief
//*----------------------------------------------------------------------------
void AT91F_CDC_Open(void)
{
    AT91C_BASE_PMC->PMC_SCER = AT91C_PMC_UDP;
    AT91C_BASE_PMC->PMC_PCER = (1 << AT91C_ID_UDP);
	
    // Set the PLL USB Divider
    AT91C_BASE_CKGR->CKGR_PLLR |= AT91C_CKGR_USBDIV_1 ;
	
	currentConfiguration = 0;
	currentConnection    = 0;
	currentRcvBank       = AT91C_UDP_RX_DATA_BK0;
	pUdp = AT91C_BASE_UDP;
	
    CDC_USB_PULLUP_PIO->PIO_OER = CDC_USB_PULLUP_MASK;
    CDC_USB_PULLUP_PIO->PIO_SODR = CDC_USB_PULLUP_MASK;
	
	debug_msg("USB CDC Initialized");
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_UDP_IsConfigured
//* \brief Test if the device is configured and handle enumeration
//*----------------------------------------------------------------------------
static uint8_t AT91F_UDP_IsConfigured(void)
{
	AT91_REG isr = pUdp->UDP_ISR;

	if (isr & AT91C_UDP_ENDBUSRES) {
		pUdp->UDP_ICR = AT91C_UDP_ENDBUSRES;
		// reset all endpoints
		pUdp->UDP_RSTEP  = (unsigned int)-1;
		pUdp->UDP_RSTEP  = 0;
		// Enable the function
		pUdp->UDP_FADDR = AT91C_UDP_FEN;
		// Configure endpoint 0
		pUdp->UDP_CSR[0] = (AT91C_UDP_EPEDS | AT91C_UDP_EPTYPE_CTRL);
	}
	else if (isr & AT91C_UDP_EPINT0) {
		pUdp->UDP_ICR = AT91C_UDP_EPINT0;
		AT91F_CDC_Enumerate();
	}
	
//	if(currentConfiguration) LED_G_ON; else LED_G_OFF;
	
	return currentConfiguration;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_UDP_Read
//* \brief Read available data from Endpoint OUT
//*----------------------------------------------------------------------------
uint32_t AT91F_UDP_Read(uint8_t *pData)
{
   uint32_t packetSize=0;
   uint32_t nbBytesRcv = 0;
   if (!AT91F_UDP_IsConfigured()) return 0;
   if (AT91C_BASE_UDP->UDP_CSR[AT91C_EP_OUT] & currentRcvBank)
   {
      packetSize = (AT91C_BASE_UDP->UDP_CSR[AT91C_EP_OUT] >> 16);
      while(packetSize--)
         pData[nbBytesRcv++] = AT91C_BASE_UDP->UDP_FDR[AT91C_EP_OUT];
      AT91C_BASE_UDP->UDP_CSR[AT91C_EP_OUT] &= ~(currentRcvBank);
      if (currentRcvBank == AT91C_UDP_RX_DATA_BK0)
         currentRcvBank = AT91C_UDP_RX_DATA_BK1;
      else
         currentRcvBank = AT91C_UDP_RX_DATA_BK0;
   }
   return nbBytesRcv;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_CDC_Write
//* \brief Send through endpoint 2
//*----------------------------------------------------------------------------
uint32_t AT91F_UDP_Write(const uint8_t *pData, uint32_t length)
{
	uint32_t cpt = 0;

	// Send the first packet
	cpt = MIN(length, AT91C_EP_IN_SIZE);
	length -= cpt;
	while (cpt--) pUdp->UDP_FDR[AT91C_EP_IN] = *pData++;
	pUdp->UDP_CSR[AT91C_EP_IN] |= AT91C_UDP_TXPKTRDY;

	while (length) {
		// Fill the second bank
		cpt = MIN(length, AT91C_EP_IN_SIZE);
		length -= cpt;
		while (cpt--) pUdp->UDP_FDR[AT91C_EP_IN] = *pData++;
		// Wait for the the first bank to be sent
		while ( !(pUdp->UDP_CSR[AT91C_EP_IN] & AT91C_UDP_TXCOMP) )
			if ( !AT91F_UDP_IsConfigured() ) return length;
		pUdp->UDP_CSR[AT91C_EP_IN] &= ~(AT91C_UDP_TXCOMP);
		while (pUdp->UDP_CSR[AT91C_EP_IN] & AT91C_UDP_TXCOMP);
		pUdp->UDP_CSR[AT91C_EP_IN] |= AT91C_UDP_TXPKTRDY;
	}
	// Wait for the end of transfer
	while ( !(pUdp->UDP_CSR[AT91C_EP_IN] & AT91C_UDP_TXCOMP) )
		if ( !AT91F_UDP_IsConfigured() ) return length;
	pUdp->UDP_CSR[AT91C_EP_IN] &= ~(AT91C_UDP_TXCOMP);
	while (pUdp->UDP_CSR[AT91C_EP_IN] & AT91C_UDP_TXCOMP);

	return length;
}


//*----------------------------------------------------------------------------
//* \fn    USB_SendData
//* \brief Send Data through the control endpoint
//*----------------------------------------------------------------------------
void USB_SendData(uint8_t *pData, uint32_t length)
{
   uint32_t cpt = 0;
   AT91_REG csr;
   uint8_t stay_in_loop=0;
   
   if((length % 8)==0) stay_in_loop = 1;
   
   do
   {
      cpt = MIN(length, 8);
      length -= cpt;
      
      while (cpt--)
      {
         (AT91C_BASE_UDP)->UDP_FDR[0] = *pData++;
      }
      
      if ((AT91C_BASE_UDP)->UDP_CSR[0] & AT91C_UDP_TXCOMP)
      {
         (AT91C_BASE_UDP)->UDP_CSR[0] &= ~(AT91C_UDP_TXCOMP);
         while ((AT91C_BASE_UDP)->UDP_CSR[0] & AT91C_UDP_TXCOMP);
      }
      
      (AT91C_BASE_UDP)->UDP_CSR[0] |= AT91C_UDP_TXPKTRDY;
      
      do
      {
         csr = (AT91C_BASE_UDP)->UDP_CSR[0];
         // Data IN stage has been stopped by a status OUT
         if (csr & AT91C_UDP_RX_DATA_BK0)
         {
            (AT91C_BASE_UDP)->UDP_CSR[0] &= ~(AT91C_UDP_RX_DATA_BK0);
            return;
         }
      } while ( !(csr & AT91C_UDP_TXCOMP) );
      
   }while(length||stay_in_loop);
   
   if ((AT91C_BASE_UDP)->UDP_CSR[0] & AT91C_UDP_TXCOMP)
   {
      (AT91C_BASE_UDP)->UDP_CSR[0] &= ~(AT91C_UDP_TXCOMP);
      while ((AT91C_BASE_UDP)->UDP_CSR[0] & AT91C_UDP_TXCOMP);
   }
}


//*----------------------------------------------------------------------------
//* \fn    AT91F_USB_SendZlp
//* \brief Send zero length packet through the control endpoint
//*----------------------------------------------------------------------------
void AT91F_USB_SendZlp(void)
{
	pUdp->UDP_CSR[0] |= AT91C_UDP_TXPKTRDY;
	while ( !(pUdp->UDP_CSR[0] & AT91C_UDP_TXCOMP) );
	pUdp->UDP_CSR[0] &= ~(AT91C_UDP_TXCOMP);
	while (pUdp->UDP_CSR[0] & AT91C_UDP_TXCOMP);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_USB_SendStall
//* \brief Stall the control endpoint
//*----------------------------------------------------------------------------
void AT91F_USB_SendStall(void)
{
	pUdp->UDP_CSR[0] |= AT91C_UDP_FORCESTALL;
	while ( !(pUdp->UDP_CSR[0] & AT91C_UDP_ISOERROR) );
	pUdp->UDP_CSR[0] &= ~(AT91C_UDP_FORCESTALL | AT91C_UDP_ISOERROR);
	while (pUdp->UDP_CSR[0] & (AT91C_UDP_FORCESTALL | AT91C_UDP_ISOERROR));
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_CDC_Enumerate
//* \brief This function is a callback invoked when a SETUP packet is received
//*----------------------------------------------------------------------------
static void AT91F_CDC_Enumerate(void)
{
	uint8_t bmRequestType, bRequest;
	uint16_t wValue, wIndex, wLength, wStatus;


	if ( !(pUdp->UDP_CSR[0] & AT91C_UDP_RXSETUP) )
		return;

	bmRequestType = pUdp->UDP_FDR[0];
	bRequest      = pUdp->UDP_FDR[0];
	wValue        = (pUdp->UDP_FDR[0] & 0xFF);
	wValue       |= (pUdp->UDP_FDR[0] << 8);
	wIndex        = (pUdp->UDP_FDR[0] & 0xFF);
	wIndex       |= (pUdp->UDP_FDR[0] << 8);
	wLength       = (pUdp->UDP_FDR[0] & 0xFF);
	wLength      |= (pUdp->UDP_FDR[0] << 8);

	if (bmRequestType & 0x80) {
		pUdp->UDP_CSR[0] |= AT91C_UDP_DIR;
		while ( !(pUdp->UDP_CSR[0] & AT91C_UDP_DIR) );
	}
	pUdp->UDP_CSR[0] &= ~AT91C_UDP_RXSETUP;
	while ( (pUdp->UDP_CSR[0]  & AT91C_UDP_RXSETUP)  );

   debug_msg("(bRequest << 8) | bmRequestType = "); debug16_t((bRequest << 8) | bmRequestType); debug_txt(" ");
	// Handle supported standard device request Cf Table 9-3 in USB specification Rev 1.1
	switch ((bRequest << 8) | bmRequestType) {
	case STD_GET_DESCRIPTOR:
      debug_txt("STD_GET_DESCRIPTOR");
		if (wValue == 0x100)       // Return Device Descriptor
      {
         debug_txt(" (device)");
			USB_SendData((uint8_t*)devDescriptor, MIN(sizeof(devDescriptor), wLength));
      }
		else if (wValue == 0x200)  // Return Configuration Descriptor
      {
         debug_txt(" (config)");
			USB_SendData((uint8_t*)cfgDescriptor, MIN(sizeof(cfgDescriptor), wLength));
      }
		else
			AT91F_USB_SendStall();
		break;
	case STD_SET_ADDRESS:
      debug_txt("STD_SET_ADDRESS, wValue = "); debug16_t(wValue);
		AT91F_USB_SendZlp();
		pUdp->UDP_FADDR = (AT91C_UDP_FEN | wValue);
		pUdp->UDP_GLBSTATE  = (wValue) ? AT91C_UDP_FADDEN : 0;
		break;
	case STD_SET_CONFIGURATION:
      debug_txt("STD_SET_CONFIGURATION, wValue = "); debug16_t(wValue);
		currentConfiguration = wValue;
		AT91F_USB_SendZlp();
		pUdp->UDP_GLBSTATE  = (wValue) ? AT91C_UDP_CONFG : AT91C_UDP_FADDEN;
		pUdp->UDP_CSR[1] = (wValue) ? (AT91C_UDP_EPEDS | AT91C_UDP_EPTYPE_BULK_OUT) : 0;
		pUdp->UDP_CSR[2] = (wValue) ? (AT91C_UDP_EPEDS | AT91C_UDP_EPTYPE_BULK_IN)  : 0;
		pUdp->UDP_CSR[3] = (wValue) ? (AT91C_UDP_EPEDS | AT91C_UDP_EPTYPE_INT_IN)   : 0;
		break;
	case STD_GET_CONFIGURATION:
      debug_txt("STD_GET_CONFIGURATION");
		USB_SendData((uint8_t *) &(currentConfiguration), sizeof(currentConfiguration));
		break;
	case STD_GET_STATUS_ZERO:
      debug_txt("STD_GET_STATUS_ZERO");
		wStatus = 0;
		USB_SendData((uint8_t *) &wStatus, sizeof(wStatus));
		break;
	case STD_GET_STATUS_INTERFACE:
      debug_txt("STD_GET_STATUS_INTERFACE");
		wStatus = 0;
		USB_SendData((uint8_t *) &wStatus, sizeof(wStatus));
		break;
	case STD_GET_STATUS_ENDPOINT:
      debug_txt("STD_GET_STATUS_INTERFACE, wIndex = "); debug16_t(wIndex);
		wStatus = 0;
		wIndex &= 0x0F;
		if ((pUdp->UDP_GLBSTATE & AT91C_UDP_CONFG) && (wIndex <= 3)) {
			wStatus = (pUdp->UDP_CSR[wIndex] & AT91C_UDP_EPEDS) ? 0 : 1;
         debug_txt(" wStatus = "); debug8_t(wStatus);
			USB_SendData((uint8_t *) &wStatus, sizeof(wStatus));
		}
		else if ((pUdp->UDP_GLBSTATE & AT91C_UDP_FADDEN) && (wIndex == 0)) {
			wStatus = (pUdp->UDP_CSR[wIndex] & AT91C_UDP_EPEDS) ? 0 : 1;
         debug_txt(" wStatus = "); debug8_t(wStatus);
			USB_SendData((uint8_t *) &wStatus, sizeof(wStatus));
		}
		else
			AT91F_USB_SendStall();
		break;
	case STD_SET_FEATURE_ZERO:
      debug_txt("STD_SET_FEATURE_ZERO, STALLed");
		AT91F_USB_SendStall();
	    break;
	case STD_SET_FEATURE_INTERFACE:
      debug_txt("STD_SET_FEATURE_INTERFACE");
		AT91F_USB_SendZlp();
		break;
	case STD_SET_FEATURE_ENDPOINT:
      debug_txt("STD_SET_FEATURE_ENDPOINT, wIndex = "); debug16_t(wIndex);
		wIndex &= 0x0F;
		if ((wValue == 0) && wIndex && (wIndex <= 3)) {
			pUdp->UDP_CSR[wIndex] = 0;
			AT91F_USB_SendZlp();
		}
		else
			AT91F_USB_SendStall();
		break;
	case STD_CLEAR_FEATURE_ZERO:
      debug_txt("STD_CLEAR_FEATURE_ZERO");
		AT91F_USB_SendStall();
	    break;
	case STD_CLEAR_FEATURE_INTERFACE:
      debug_txt("STD_CLEAR_FEATURE_INTERFACE");
		AT91F_USB_SendZlp();
		break;
	case STD_CLEAR_FEATURE_ENDPOINT:
      debug_txt("STD_CLEAR_FEATURE_ENDPOINT, wIndex = "); debug16_t(wIndex); debug_txt(", wValue = "); debug16_t(wValue);
		wIndex &= 0x0F;
		if ((wValue == 0) && wIndex && (wIndex <= 3)) {
			if (wIndex == 1)
				pUdp->UDP_CSR[1] = (AT91C_UDP_EPEDS | AT91C_UDP_EPTYPE_BULK_OUT);
			else if (wIndex == 2)
				pUdp->UDP_CSR[2] = (AT91C_UDP_EPEDS | AT91C_UDP_EPTYPE_BULK_IN);
			else if (wIndex == 3)
				pUdp->UDP_CSR[3] = (AT91C_UDP_EPEDS | AT91C_UDP_EPTYPE_ISO_IN);
			AT91F_USB_SendZlp();
		}
		else
			AT91F_USB_SendStall();
		break;

	// handle CDC class requests
	case SET_LINE_CODING:
		while ( !(pUdp->UDP_CSR[0] & AT91C_UDP_RX_DATA_BK0) );
		debug_msg("SET_LINE_CODING, packet: ");
      
		//read data
		uint16_t i;
		uint8_t *buf;
		buf = (void*)&line;
		for(i=0;i<wLength;i++)
      {
         buf[i] = AT91C_BASE_UDP->UDP_FDR[0];
         debug8_t(buf[i]); debug_chr(' ');
      }
		
		//confirm receipt in UDP
		pUdp->UDP_CSR[0] &= ~(AT91C_UDP_RX_DATA_BK0);
		
		//set the flag (inform a main loop or CDC task)
		lineCodingChanged=1;
		
		//response for host
		AT91F_USB_SendZlp();
		break;
	case GET_LINE_CODING:
      debug_txt("GET_LINE_CODING");
		USB_SendData((uint8_t *) &line, MIN(sizeof(line), wLength));
		break;
	case SET_CONTROL_LINE_STATE:
      debug_txt("SET_CONTROL_LINE_STATE, wValue = "); debug16_t(wValue);
		currentConnection = wValue;
		AT91F_USB_SendZlp();
		break;
	default:
		AT91F_USB_SendStall();
	    break;
	}
}
