#include "common.h"


void init(void)
{
	// enable peripheral clock for PIO controllers
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA);
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
	
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	//open debug channel
	debug_init_default();
	
	//initialize LEDs
	LED_R_INIT;
	LED_Y_INIT;
	LED_G_INIT;
	LED_B_INIT;
	
}

#define SPI_PINS_MASK \
AT91C_PA12_SPI0_NPCS0 | \
AT91C_PA13_SPI0_NPCS1 | \
AT91C_PA14_SPI0_NPCS2 | \
AT91C_PA15_SPI0_NPCS3 | \
AT91C_PA16_SPI0_MISO | \
AT91C_PA17_SPI0_MOSI | \
AT91C_PA18_SPI0_SPCK

#define SPI_PIO				AT91C_BASE_PIOA

#define SPI					AT91C_BASE_SPI0
#define SPI_ID				AT91C_ID_SPI0

const uint32_t NPCS_CODE_MASK[4] = { 0<<16, 1<<16, 3<<16, 7<<16 };



uint16_t spiTransfer(uint16_t dataToSend, uint8_t channel)
{
	uint16_t receivedData;
	
	while( !( SPI->SPI_SR & AT91C_SPI_TDRE ) );
	SPI->SPI_TDR = NPCS_CODE_MASK[channel] | dataToSend;
	
	while( !( SPI->SPI_SR & AT91C_SPI_RDRF ) );
	receivedData = (uint16_t)( SPI->SPI_RDR );
	
	return(receivedData);
}



void spiDmaTransfer(void *buffer, uint16_t len, uint8_t channel)
{
	//make a copy of SPI_MR
	uint32_t SPI_MR_backup = SPI->SPI_MR;
	
	//make a copy of SPI_CSR[0]
	uint32_t SPI_CSR0_backup = SPI->SPI_CSR[0];
	
	//clear PCS field and variable peripheral selection
	SPI->SPI_MR &= ~(AT91C_SPI_PCS | AT91C_SPI_PS);
	
	//activate desired channel
	SPI->SPI_MR |= NPCS_CODE_MASK[channel];
	
	//clear BITS field in CSR0 register
	SPI->SPI_CSR[0] &= ~AT91C_SPI_BITS;
	
	//copy BITS field from channel's CSR to CSR0
	SPI->SPI_CSR[0] |= SPI->SPI_CSR[channel] & AT91C_SPI_BITS;
	
	//prepare DMA transfer
	SPI->SPI_TPR = (uint32_t)buffer;
	SPI->SPI_TNCR = 0;
	SPI->SPI_TCR = len;
	
	//start DMA transfer
	SPI->SPI_PTCR = AT91C_PDC_TXTEN;
	
	//wait for end of transfer
	while( !( SPI->SPI_SR & AT91C_SPI_TXEMPTY ) );
	
	//restore previous SPI_MR state
	SPI->SPI_MR  = SPI_MR_backup;
	
	//restore previous SPI_CSR0 state
	SPI->SPI_CSR[0] = SPI_CSR0_backup;
	
	//dummy read clears RDRF flag
	volatile uint8_t dummy;
	dummy = SPI->SPI_RDR;
}


int main( void )
{
	init();
	
	//initialize SPI (common for all channels)
	
	//disable PIO control on hardware controlled SPI pins
	SPI_PIO->PIO_PDR = SPI_PINS_MASK;
	
	//SPI pins are peripheral A
	SPI_PIO->PIO_ASR = SPI_PINS_MASK;
	
	//enable SPI clock
	AT91C_BASE_PMC->PMC_PCER = ( 1 << SPI_ID );
	
	//SPI enable and reset
	SPI->SPI_CR = AT91C_SPI_SWRST;
	
	const uint8_t DELAY_BETWEEN_CHIP_SELECTS = 0xFF;
	const uint32_t DLYBCS_MASK	= (uint32_t)DELAY_BETWEEN_CHIP_SELECTS << 24;
	
	//SPI Mode Register: Fault detection disable, variable periph. chip select
    SPI->SPI_MR  = DLYBCS_MASK | AT91C_SPI_MSTR | AT91C_SPI_MODFDIS | AT91C_SPI_PS;
	
	SPI->SPI_CR = AT91C_SPI_SPIEN;
	
	const uint8_t BAUD_DIVIDER = 48;
	const uint8_t DELAY_BEFORE_SPCK = 10;
	const uint8_t DELAY_BETWEEN_CONSECUTIVE_TRANSFERS = 0;	
	
	//masks for SPI_CSR registers
	const uint32_t SCBR_MASK = (uint32_t)BAUD_DIVIDER << 8;
	const uint32_t DLYBS_MASK = (uint32_t)DELAY_BEFORE_SPCK << 16;
	const uint32_t DLYBCT_MASK = (uint32_t)DELAY_BETWEEN_CONSECUTIVE_TRANSFERS << 24;
	
	//initialize NPCS0 to "Mode 0"
	SPI->SPI_CSR[0] = AT91C_SPI_BITS_8 | SCBR_MASK | DLYBS_MASK | DLYBCT_MASK | AT91C_SPI_NCPHA;
	
	//initialize NPCS1 to "Mode 1"
	SPI->SPI_CSR[1] = AT91C_SPI_BITS_8 | SCBR_MASK | DLYBS_MASK | DLYBCT_MASK;
	
	//initialize NPCS2 to "Mode 2"
	SPI->SPI_CSR[2] = AT91C_SPI_BITS_8 | SCBR_MASK | DLYBS_MASK | DLYBCT_MASK | AT91C_SPI_NCPHA | AT91C_SPI_CPOL;
	
	//initialize NPCS3 to "Mode 3"
	SPI->SPI_CSR[3] = AT91C_SPI_BITS_8 | SCBR_MASK | DLYBS_MASK | DLYBCT_MASK | AT91C_SPI_CPOL;
	
	
	for(;;)
	{
		uint8_t key;
		uint8_t buffer[3] = {0x55, 0x00, 0xFF};
		
		key = debug_inkey();
		
		switch(key)
		{
			case '1':
			{
				debug_msg("0x55 on NPCS0 in Mode 0");
				spiTransfer(0x55, 0);
				break;
			}
			case '2':
			{
				debug_msg("0x55 on NPCS1 in Mode 1");
				spiTransfer(0x55, 1);
				break;
			}
			case '3':
			{
				debug_msg("0x55 on NPCS2 in Mode 2");
				spiTransfer(0x55, 2);
				break;
			}
			case '4':
			{
				debug_msg("0x55 on NPCS3 in Mode 3");
				spiTransfer(0x55, 3);
				break;
			}
			case 'q':
			{
				debug_msg("DMA: 0x55 0x00 0xFF on NPCS0");
				spiDmaTransfer(buffer,3,0);
				break;
			}
			case 'w':
			{
				debug_msg("DMA: 0x55 0x00 0xFF on NPCS1");
				spiDmaTransfer(buffer,3,1);
				break;
			}
			case 'e':
			{
				debug_msg("DMA: 0x55 0x00 0xFF on NPCS2");
				spiDmaTransfer(buffer,3,2);
				break;
			}
			case 'r':
			{
				debug_msg("DMA: 0x55 0x00 0xFF on NPCS3");
				spiDmaTransfer(buffer,3,3);
				break;
			}
			case 'a':
			{
				debug_msg("0x55 0x00 0xFF on NPCS0");
				spiTransfer(0x55,0);
				spiTransfer(0x00,0);
				spiTransfer(0xFF,0);
				break;
			}
			case 's':
			{
				debug_msg("0x55 0x00 0xFF on NPCS1");
				spiTransfer(0x55,1);
				spiTransfer(0x00,1);
				spiTransfer(0xFF,1);
				break;
			}
			case 'd':
			{
				debug_msg("0x55 0x00 0xFF on NPCS2");
				spiTransfer(0x55,2);
				spiTransfer(0x00,2);
				spiTransfer(0xFF,2);
				break;
			}
			case 'f':
			{
				debug_msg("0x55 0x00 0xFF on NPCS3");
				spiTransfer(0x55,3);
				spiTransfer(0x00,3);
				spiTransfer(0xFF,3);
				break;
			}
			case 'z':
			{
				debug_msg("0x55 on NPCS0, 0x00 on NPCS1, 0xFF on NPCS2");
				spiTransfer(0x55,0);
				spiTransfer(0x00,1);
				spiTransfer(0xFF,2);
				break;
			}
			case 'x':
			{
				debug_msg("generating erroneous number of SPCK pulses on NPCS1");
				//backup spi CSR register
				uint32_t SPI_CSR_Backup = SPI->SPI_CSR[1];
				
				//clear CSR register
				SPI->SPI_CSR[1] = 0;
				
				//set baudrate to max (SBBR field = 1), 9 bits for each transfer
				const uint32_t MAX_SCBR_MASK = (1<<8);
				SPI->SPI_CSR[1] = MAX_SCBR_MASK | AT91C_SPI_BITS_9;
				
				spiTransfer(0x155,1);
				
				//restore spi CSR register
				SPI->SPI_CSR[1] = SPI_CSR_Backup;
			}
		}//switch(key)
		
	}
	return(0);
}


